"""Builder for tables from flat files for valueset lookup"""

import csv
import dataclasses
import pathlib

import pandas

from cumulus_library import BaseTableBuilder, base_utils, study_manifest
from cumulus_library.builders.valueset import valueset_utils, vsac
from cumulus_library.template_sql import base_templates


@dataclasses.dataclass(kw_only=True)
class TableConfig:
    """Convenience class for holding params for configuring tables from flat files"""

    file_path: str
    delimiter: str
    table_name: str
    headers: list[str]
    dtypes: dict
    parquet_types: list[str]
    filtered_path: str | None = None
    map_cols: list[dict] | None = None


class StaticBuilder(BaseTableBuilder):
    def get_keywords_table_configs(
        self,
        config: base_utils.StudyConfig,
        manifest: study_manifest.StudyManifest,
        prefix: str | None = None,
    ):
        if prefix is None:
            prefix = ""  # pragma: no cover
        configs = []
        # Do we have a keyword lookup? if so, upload it and the associated rules file
        if self.valueset_config.keyword_file is not None:
            configs += [
                TableConfig(
                    file_path=self.study_path / self.valueset_config.keyword_file,
                    delimiter="\t",
                    table_name=f"{prefix}keywords",
                    headers=["STR"],
                    dtypes={"STR": "str"},
                    parquet_types=["STRING"],
                    filtered_path=self.data_path / "./keywords.filtered.tsv",
                )
            ]
        else:
            self.queries.append(
                base_templates.get_ctas_empty_query(
                    schema_name=config.schema,
                    table_name=f"{manifest.get_study_prefix()}__{prefix}keywords",
                    table_cols=["STR"],
                )
            )

        # This sets the source of the expansion rules, or uses our reasonable default
        # TODO: Move this in with the UMLS generation
        if self.valueset_config.rules_file:
            rules_path = self.study_path / self.valueset_config.rules_file
        else:
            rules_path = (
                pathlib.Path(__file__).resolve().parents[0]
                / "lookup_drug_from_ingredient_rules.tsv"
            )
        configs += [
            TableConfig(
                file_path=rules_path,
                delimiter="\t",
                table_name=f"{prefix}search_rules",
                headers=[
                    "TTY1",
                    "RELA",
                    "TTY2",
                    "rule",
                ],
                dtypes={"TTY1": "str", "RELA": "str", "TTY2": "str", "rule": "str"},
                parquet_types=["STRING", "STRING", "STRING", "STRING", "BOOLEAN"],
                map_cols=[
                    {
                        "from": "rule",
                        "to": "include",
                        "map_dict": {"yes": True, "no": False},
                    }
                ],
            )
        ]
        return configs

    def filter_duplicate_entries(
        self, path: pathlib.Path, delimiter: str, filtered_path: pathlib.Path
    ):
        """Given a dataset, returns the set of shortest unique substrings from that set

        As an example, given a dataset like this:
            'Hydone Formula Liquid'
            'Hydone Formula Liquid Oral Liquid Product'
            'Hydone Formula Liquid Oral Product'
            'Novahistine Expectorant'
            'Novahistine Expectorant, 10 mg-100 mg-30 mg/5 mL oral liquid'
            'Novahistine Expectorant Oral Liquid Product'
            'Novahistine Expectorant Oral Product'
            'Codeine'
            'aconite / codeine / Erysimum preparation'
            'codeine / potassium'

        This function should return:
            'Codiene'
            'Hydone Formula Liquid'
            'Novahistine Expectorant'

        :param path: path to a file (expected to be two cols, with data in second col)
        :param delimiter: the character used to separate columns in that file
        :param filtered_path: The name of the file to write the filtered dataset to

        """
        target = self.data_path / f"{path.stem}.filtered.tsv"
        with open(path) as file:
            reader = csv.reader(file, delimiter=delimiter)
            keywords = sorted((row[0] for row in reader), key=len)
        unique_keywords = {}  # casefold -> normal
        for keyword in keywords:
            folded = keyword.casefold()
            if not any(x in folded for x in unique_keywords):
                unique_keywords[folded] = keyword

        df = pandas.DataFrame(unique_keywords.values())
        df.to_csv(target, sep="\t", index=False, header=False)

    def prepare_queries(
        self,
        *args,
        config: base_utils.StudyConfig,
        manifest: study_manifest.StudyManifest,
        valueset_config: valueset_utils.ValuesetConfig,
        toml_path: pathlib.Path,
        **kwargs,
    ):
        self.valueset_config = valueset_config
        self.study_path = manifest._study_path
        self.data_path = valueset_utils.get_valueset_cache_dir(toml_path, None)
        prefix = self.valueset_config.table_prefix
        if prefix:
            prefix += "_"

        self.tables = self.get_keywords_table_configs(config, manifest, prefix)

        # For each VSAC dataset, we'll get the topics defined by that dataset
        vsac_df = pandas.DataFrame(columns=["sab", "rxcui", "display", "steward", "oid"])
        for key in valueset_config.vsac_stewards:
            vsac.download_oid_data(
                steward=key,
                oid=valueset_config.vsac_stewards[key],
                config=config,
                manifest=None,
                path=toml_path,
            )
            steward_df = pandas.read_csv(
                self.data_path / f"{key}.tsv",
                delimiter="\t",
                names=["sab", "rxcui", "display"],
                header=0,
            )

            steward_df["steward"] = key
            steward_df["oid"] = valueset_config.vsac_stewards[key]
            vsac_df = pandas.concat([vsac_df, steward_df])
            vsac_df["rxcui"] = vsac_df["rxcui"].astype("str")
        vsac_df.to_csv(self.data_path / "all_vsac.tsv", sep="\t", index=False)

        # And now we'll add the VSAC static table to the list of tables to upload
        self.tables.append(
            TableConfig(
                file_path=self.data_path / "all_vsac.tsv",
                delimiter="\t",
                table_name=f"{prefix}vsac_valuesets",
                headers=["sab", "rxcui", "str", "steward", "oid"],
                dtypes={
                    "sab": "str",
                    "rxcui": "str",
                    "display": "str",
                    "steward": "str",
                    "oid": "str",
                },
                parquet_types=["STRING", "STRING", "STRING", "STRING", "STRING"],
            )
        )

        with base_utils.get_progress_bar() as progress:
            task = progress.add_task("Uploading static files...", total=len(self.tables))

            for table in self.tables:
                # Determine what we're using as a source file
                if table.filtered_path:
                    self.filter_duplicate_entries(
                        table.file_path, table.delimiter, table.filtered_path
                    )
                    path = self.data_path / table.filtered_path
                else:
                    path = self.data_path / table.file_path
                parquet_path = path.with_suffix(".parquet")

                # Read the file, using lots of the TableConfig params, and generate
                # a parquet file

                df = pandas.read_csv(
                    path,
                    delimiter=table.delimiter,
                    names=table.headers,
                    header=0,
                    dtype=table.dtypes,
                    index_col=False,
                    na_values=["\\N"],
                )
                if table.map_cols:
                    for mapping in table.map_cols:
                        df[mapping["to"]] = df[mapping["from"]].str.lower().map(mapping["map_dict"])
                        table.headers.append(mapping["to"])
                df.to_parquet(parquet_path)
                # Upload to S3 and create a table that reads from it
                prefix = manifest.get_study_prefix()
                remote_path = config.db.upload_file(
                    file=parquet_path,
                    study=prefix,
                    topic=parquet_path.stem,
                    force_upload=config.force_upload,
                )
                self.queries.append(
                    base_templates.get_ctas_from_parquet_query(
                        schema_name=config.schema,
                        table_name=f"{prefix}__{table.table_name}",
                        local_location=parquet_path,
                        remote_location=remote_path,
                        table_cols=table.headers,
                        remote_table_cols_types=table.parquet_types,
                    )
                )
                progress.advance(task)



import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "allergyintolerance": {
        "id": [],
        "type": [],
        "category": [],
        "criticality": [],
        "recordedDate": [],
        "patient": sql_utils.REFERENCE,
        "encounter": sql_utils.REFERENCE,
        "reaction": ["severity"],
    }
}


class CoreAllergyIntoleranceBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating AllergyIntolerance tables..."

    def prepare_queries(self, *args, config: cumulus_library.StudyConfig, **kwargs):
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries.append(
            core_templates.get_core_template("allergyintolerance", validated_schema)
        )



import cumulus_library
from cumulus_library.template_sql import sql_utils


class CoreDocumentreferenceBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating DocumentReference tables..."

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        self.queries = sql_utils.denormalize_complex_objects(
            config.db,
            [
                sql_utils.CodeableConceptConfig(
                    source_table="documentreference",
                    source_id="id",
                    column_hierarchy=[("type", dict)],
                    target_table="core__documentreference_dn_type",
                ),
                sql_utils.CodeableConceptConfig(
                    source_table="documentreference",
                    source_id="id",
                    column_hierarchy=[("category", list)],
                    target_table="core__documentreference_dn_category",
                ),
                sql_utils.CodingConfig(
                    source_table="documentreference",
                    source_id="id",
                    column_hierarchy=[("content", list), ("format", dict)],
                    target_table="core__documentreference_dn_format",
                    expected={"format": sql_utils.CODING},
                ),
            ],
            "DocumentReference",
        )



"""Module for extracting US core extensions from patient records"""

import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "observation": {
        "id": [],
        "component": {
            "valueQuantity": ["code", "comparator", "system", "unit", "value"],
        },
        "status": [],
        "effectiveDateTime": [],
        "valueQuantity": ["code", "comparator", "system", "unit", "value"],
        "valueString": [],
        "subject": sql_utils.REFERENCE,
        "encounter": sql_utils.REFERENCE,
    }
}


class ObservationBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating Observation tables..."

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        """constructs queries related to patient extensions of interest

        :param config: A study config object
        """
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries += [
            core_templates.get_core_template("observation", validated_schema),
            core_templates.get_core_template(
                "observation_component_valuequantity", validated_schema
            ),
        ]



# Variables in this file should only be written to by databases.py, and should be
# treated as read only elsewhere (and really only by jinja envrionments)
# TODO: Consider moving to config object passed down from main cli to builders
db_type = None



"""Creates medication lists from UMLS

We're leveraging the MRCONSO and MRREL tables to do some recursive lookups.
For a given medical language system, we'll build up some relationships,
and then use those as a seed for the next iteration, continuing until we've
identified all concepts related to a specific set of seed values.

Source terminology (see UMLS docs for more details):

https://www.ncbi.nlm.nih.gov/books/NBK9685/table/ch03.T.concept_names_and_sources_file_mr/
MRCONSO (source of concept information):
    SAB: Abbreviated source name of the concept. This is the name of the source
        medical vocabulary, i.e. SNOMED, ICD10, MEDRT, etc
    CUI: the unique identifier for a concept in UMLS
    TTY: Term type abbreviation, i.e. CD for clinical drug, DF for dose form, etc
    CODE: The primary ID for a concept in the source vocabulary
    STR: The display name of the concept, like a drug's name

https://www.ncbi.nlm.nih.gov/books/NBK9685/table/ch03.T.related_concepts_file_mrrel_rrf/
MRREL (source of mapping between concepts):
    SAB: Abbreviated source name of the relationship. Same as MRCONSO SAB
    CUI1: The UMLS ID of the concept you're mapping from
    CUI2: The UMLS ID of the concept you're mapping to
    REL: An abbreviation mapping to a relationship type, like RB (broader relationship),
        or RN (narrower relationship)
    RELA: The relationship between the concepts, like `tradename_of` or
        `has_ingredient`. May be null.
"""

import pathlib

from cumulus_library import base_utils, study_manifest
from cumulus_library.builders.valueset import valueset_utils
from cumulus_library.template_sql import base_templates


def generate_umls_tables(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    valueset_config: valueset_utils.ValuesetConfig,
):
    base_path = pathlib.Path(__file__).resolve().parent
    study_prefix = manifest.get_prefix_with_seperator()
    cursor = config.db.cursor()
    table_prefix = ""
    if valueset_config.table_prefix:
        table_prefix = valueset_config.table_prefix + "_"
    create_table = True
    cursor.execute(f"DROP TABLE IF EXISTS {study_prefix}{table_prefix}valueset_rels")
    cursor.execute(f"DROP TABLE IF EXISTS {study_prefix}{table_prefix}valueset")
    # For each UMLS steward defined, we'll iterate through UMLS, using CUIs to follow links
    # through the UMLS tree, excluding CUIs we've already found.
    for steward in valueset_config.umls_stewards:
        cursor = config.db.cursor()
        display_text = f"Discovering {steward} medications, pass "
        tier = 1
        with base_utils.get_progress_bar(disable=config.verbose) as progress:
            task = progress.add_task(
                display_text + str(tier), total=None, visible=not config.verbose
            )
            query = base_templates.get_template(
                "umls_iterate",
                base_path / "template_sql",
                steward=steward,
                prefix=study_prefix,
                table_prefix=table_prefix,
                sab=valueset_config.umls_stewards[steward]["sab"],
                search_terms=valueset_config.umls_stewards[steward]["search_terms"],
                tier=tier,
                create_table=create_table,
            )
            create_table = False
            cursor.execute(query)
            prev = 0
            # our break condition is if we find no new rows after an iteration
            current = cursor.execute(
                f"SELECT count(*) from {study_prefix}{table_prefix}umls_valuesets_rels"  # noqa: S608
            ).fetchone()[0]
            while current != prev:
                prev = current
                tier += 1
                progress.update(task, description=display_text + str(tier))
                query = base_templates.get_template(
                    "umls_iterate",
                    base_path / "template_sql",
                    steward=steward,
                    prefix=study_prefix,
                    table_prefix=table_prefix,
                    sab=valueset_config.umls_stewards[steward]["sab"],
                    search_terms=valueset_config.umls_stewards[steward]["search_terms"],
                    tier=tier,
                )
                cursor.execute(query)
                current = cursor.execute(
                    f"SELECT count(*) from {study_prefix}{table_prefix}umls_valuesets_rels"  # noqa: S608
                ).fetchone()[0]
    rels_cols = [
        "steward",
        "tier",
        "cui",
        "cui1",
        "rel",
        "rela",
        "cui2",
        "sab",
        "tty",
        "code",
        "str",
    ]
    # with the iteration complete, we'll grab the CUI of the actual topic, and
    # append it to the rel so we have a line to its parents and children
    if valueset_config.umls_stewards:
        query = base_templates.get_create_table_from_tables(
            table_name=f"{study_prefix}{table_prefix}umls_valuesets",
            tables=[
                f"{study_prefix}{table_prefix}umls_valuesets_rels",
                # TODO: allow this to target multiple codings for joins
                "rxnorm.RXNCONSO",
            ],
            columns=["r.rxcui"] + [f"v.{x}" for x in rels_cols],
            join_clauses=["v.code = r.code", "v.sab = r.sab"],
            table_aliases=["v", "r"],
            distinct=True,
        )
        cursor.execute(query)
    # Or, just create an empty table if no UMLS stewards were defined
    else:
        query = base_templates.get_ctas_empty_query(
            schema_name=config.schema,
            table_name=f"{study_prefix}{table_prefix}umls_valuesets",
            table_cols=["rxcui", *rels_cols],
        )
        cursor.execute(query)



"""Module for extracting US core extensions from patient records"""

from dataclasses import dataclass

import cumulus_library
from cumulus_library.template_sql import sql_utils


@dataclass(kw_only=True)
class ObsConfig(sql_utils.CodeableConceptConfig):
    source_table: str = "observation"
    is_public: bool = False

    def __post_init__(self):
        # Consideration for future: should all denormalized tables be public?
        # For now, we'll mark the ones we want to encourage use of,
        # and for those, remove the maybe-confusing denormalization tag.
        table_suffix = "" if self.is_public else "dn_"
        table_suffix += "_".join(c[0] for c in self.column_hierarchy)
        self.target_table = f"core__observation_{table_suffix}"


class ObservationBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating Observation tables..."

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        """constructs queries related to patient extensions of interest

        :param config: A study config object
        """
        code_sources = [
            ObsConfig(column_hierarchy=[("category", list)], filter_priority=False),
            ObsConfig(column_hierarchy=[("code", dict)], filter_priority=False),
            ObsConfig(
                is_public=True,
                column_hierarchy=[("component", list), ("code", dict)],
                expected={"code": sql_utils.CODEABLE_CONCEPT},
            ),
            ObsConfig(
                is_public=True,
                column_hierarchy=[("component", list), ("dataabsentreason", dict)],
                expected={"dataabsentreason": sql_utils.CODEABLE_CONCEPT},
            ),
            ObsConfig(
                is_public=True,
                column_hierarchy=[("component", list), ("interpretation", list)],
                expected={"interpretation": sql_utils.CODEABLE_CONCEPT},
            ),
            ObsConfig(
                is_public=True,
                column_hierarchy=[("component", list), ("valuecodeableconcept", dict)],
                expected={"valuecodeableconcept": sql_utils.CODEABLE_CONCEPT},
            ),
            ObsConfig(column_hierarchy=[("interpretation", list)], filter_priority=False),
            ObsConfig(
                column_hierarchy=[("valuecodeableconcept", dict)],
                filter_priority=False,
            ),
            ObsConfig(
                column_hierarchy=[("dataabsentreason", dict)],
                filter_priority=False,
            ),
        ]

        self.queries += sql_utils.denormalize_complex_objects(
            config.db, code_sources, "Observation"
        )



"""A list of files/filename partials to be allowed to be uploaded to an aggregator

While a valid generation target, `.archive.` should never be part of this list, since
that explicitly contains line level data.
"""

ALLOWED_UPLOADS = [".cube.", ".annotated_cube.", ".flat.", ".meta.", "manifest.toml"]



"""Builder for creating tables for tracking state/logging changes"""

import pathlib
import tomllib

from cumulus_library import (
    BaseTableBuilder,
    base_utils,
    enums,
    study_manifest,
)
from cumulus_library.template_sql import base_templates

TRANSACTIONS_COLS = ["study_name", "library_version", "status", "event_time", "message"]
TRANSACTION_COLS_TYPES = ["varchar", "varchar", "varchar", "timestamp", "varchar"]
# while it may seem redundant, study_name and view_name are included as a column for
# ease of constructing a view of multiple transaction tables
STATISTICS_COLS = [
    "study_name",
    "library_version",
    "table_type",
    "table_name",
    "view_name",
    "created_on",
]
STATISTICS_COLS_TYPES = [
    "varchar",
    "varchar",
    "varchar",
    "varchar",
    "varchar",
    "timestamp",
]


class ProtectedTableBuilder(BaseTableBuilder):
    """Builder for tables that persist across study clean/build actions"""

    display_text = "Creating/updating system tables..."

    def prepare_queries(
        self,
        config: base_utils.StudyConfig,
        manifest: study_manifest.StudyManifest | None = None,
        *args,
        study_stats: dict | None = None,
        **kwargs,
    ):
        study_stats = study_stats or {}
        if manifest and manifest.get_dedicated_schema():
            db_schema = manifest.get_dedicated_schema()
            transactions = enums.ProtectedTables.TRANSACTIONS.value
            statistics = enums.ProtectedTables.STATISTICS.value
        else:
            db_schema = config.schema
            transactions = (
                f"{manifest.get_study_prefix()}__{enums.ProtectedTables.TRANSACTIONS.value}"
            )
            statistics = f"{manifest.get_study_prefix()}__{enums.ProtectedTables.STATISTICS.value}"
        self.queries.append(
            base_templates.get_ctas_empty_query(
                db_schema,
                transactions,
                TRANSACTIONS_COLS,
                TRANSACTION_COLS_TYPES,
            )
        )
        files = manifest.get_all_workflows()
        if len(files) == 0:
            return
        stats_types = set(item.value for item in enums.StatisticsTypes)
        # In this loop, we are just checking to see if :any: workflow is a stats
        # type workflow - if so, we'll create a table to hold data of stats runs
        # (if it doesn't already exist) outside of the study lifecycle for
        # persistence reasons
        for file in files:
            toml_path = pathlib.Path(f"{manifest._study_path}/{file}")
            with open(toml_path, "rb") as file:
                workflow_config = tomllib.load(file)
                if workflow_config["config_type"] in stats_types:
                    self.queries.append(
                        base_templates.get_ctas_empty_query(
                            db_schema,
                            statistics,
                            STATISTICS_COLS,
                            STATISTICS_COLS_TYPES,
                        )
                    )
                    return
